#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <inttypes.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <pthread.h>
#include <sys/utsname.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/uio.h>

#include <xs.h>
#include "xennerctrl.h"
#include <xen/io/xenbus.h>

#include "blkif.h"
#include "shared.h"
#include "list.h"
#include "hexdump.h"
#include "daemon.h"
#include "xenbackd.h"
#include "evtchnd.h"

/* ------------------------------------------------------------- */

#define BE_ROOT     "/local/domain/0/backend/blkbackd"
#define BLOCK_SIZE  512

enum blkif_state {
    BLKIF_UNKNOWN = 0,
    BLKIF_PROBED,
    BLKIF_INITIALISED,
    BLKIF_CONNECTED,
    BLKIF_DISCONNECTED,
};

struct ioreq {
    blkif_request_t     req;
    int16_t             status;

    /* parsed request */
    off_t               start, end;
    struct iovec        vec[BLKIF_MAX_SEGMENTS_PER_REQUEST];
    int                 vecs;
    int                 presync;
    int                 postsync;

    struct blkdev       *blkdev;
    struct list_head    list;
};

struct blkdev {
    struct xendev       xendev;  /* must be first */
    char                *params;
    char                *mode;
    int                 ring_ref;
    int                 file;
    size_t              file_blk;
    size_t              file_size;
    int                 protocol;
    blkif_back_rings_t  rings;
    int                 more_work;
    enum blkif_state    state;

    /* worker thread */
    pthread_t           mainthread;
    pthread_cond_t      waiter;
    pthread_mutex_t     waitlock;
    
    /* request lists */
    pthread_mutex_t     listlock;
    struct list_head    inflight;
    struct list_head    finished;
    struct list_head    freelist;
    int                 requests;

    /* statistics */
    uint32_t            req_read;      /* read requests          */
    uint32_t            req_write;     /* write requests (incl barriers) */
    uint32_t            req_barrier;   /* barrier writes         */
    uint32_t            req_full;      /* # of queue full events */
    uint32_t            req_err_parse; /* parse errors           */
    uint32_t            req_err_io;    /* I/O errors             */
    uint64_t            bytes_read;    /* bytes read             */
    uint64_t            bytes_write;   /* bytes written          */
};

static int insecure;
static int syncwrite;
static int use_threads  = 1;
static int max_requests = 32;

static pthread_attr_t joinable;
static pthread_attr_t detached;

/* ------------------------------------------------------------- */

static struct ioreq *ioreq_start(struct blkdev *blkdev)
{
    struct ioreq *ioreq = NULL;

    pthread_mutex_lock(&blkdev->listlock);
    if (list_empty(&blkdev->freelist)) {
	if (blkdev->requests >= max_requests) {
	    d1printf("%s: request limit (%d) reached\n", __FUNCTION__, max_requests);
	    blkdev->req_full++;
	    goto unlock;
	}
	/* allocate new struct */
	ioreq = malloc(sizeof(*ioreq));
	if (NULL == ioreq)
	    goto unlock;
	memset(ioreq, 0, sizeof(*ioreq));
	ioreq->blkdev = blkdev;
	blkdev->requests++;
    } else {
	/* get one from freelist */
	ioreq = list_entry(blkdev->freelist.next, struct ioreq, list);
	list_del(&ioreq->list);
    }
    list_add_tail(&ioreq->list, &blkdev->inflight);

unlock:
    pthread_mutex_unlock(&blkdev->listlock);
    return ioreq;
}

static void ioreq_finish(struct ioreq *ioreq)
{
    struct blkdev *blkdev = ioreq->blkdev;

    pthread_mutex_lock(&blkdev->listlock);
    list_del(&ioreq->list);
    list_add_tail(&ioreq->list, &blkdev->finished);
    pthread_mutex_unlock(&blkdev->listlock);
}

static void __ioreq_release(struct ioreq *ioreq)
{
    struct blkdev *blkdev = ioreq->blkdev;

    list_del(&ioreq->list);
    memset(ioreq, 0, sizeof(*ioreq));
    ioreq->blkdev = blkdev;
    list_add_tail(&ioreq->list, &blkdev->freelist);
}

static void ioreq_release(struct ioreq *ioreq)
{
    pthread_mutex_lock(&ioreq->blkdev->listlock);
    __ioreq_release(ioreq);
    pthread_mutex_unlock(&ioreq->blkdev->listlock);
}

/*
 * translate request into iovec + start offset + end offset
 * do sanity checks along the way
 */
static int ioreq_parse(struct ioreq *ioreq)
{
    struct blkdev *blkdev = ioreq->blkdev;
    void *page, *mem;
    size_t len;
    int i;

    d2printf("%s: op %d, nr %d, handle %d, id %" PRId64 ", sector %" PRId64 "\n",
	     __FUNCTION__, ioreq->req.operation, ioreq->req.nr_segments,
	     ioreq->req.handle, ioreq->req.id, ioreq->req.sector_number);
    switch (ioreq->req.operation) {
    case BLKIF_OP_READ:
	blkdev->req_read++;
	if (BLKIF_OP_READ != ioreq->req.operation && blkdev->mode[0] != 'w') {
	    d1printf("%s: error: write req for ro device\n", __FUNCTION__);
	    goto err;
	}
    case BLKIF_OP_WRITE_BARRIER:
	blkdev->req_barrier++;
	if (!syncwrite)
	    ioreq->presync = ioreq->postsync = 1;
	/* fall through */
    case BLKIF_OP_WRITE:
	blkdev->req_write++;
	if (syncwrite)
	    ioreq->postsync = 1;
	break;
    };
    
    ioreq->start = ioreq->end = ioreq->req.sector_number * blkdev->file_blk;
    for (i = 0; i < ioreq->req.nr_segments; i++) {
	if (i == BLKIF_MAX_SEGMENTS_PER_REQUEST) {
	    d1printf("%s: error: nr_requests too big\n", __FUNCTION__);
	    goto err;
	}
	d2printf("%s:   %d: gref %d sects %d-%d\n", __FUNCTION__,
		 i, ioreq->req.seg[i].gref, ioreq->req.seg[i].first_sect, ioreq->req.seg[i].last_sect);
	if (ioreq->req.seg[i].first_sect > ioreq->req.seg[i].last_sect) {
	    d1printf("%s: first > last sector\n", __FUNCTION__);
	    goto err;
	}
	if (ioreq->req.seg[i].last_sect * BLOCK_SIZE >= PAGE_SIZE) {
	    d1printf("%s: page crossing\n", __FUNCTION__);
	    goto err;
	}
	len = (ioreq->req.seg[i].last_sect - ioreq->req.seg[i].first_sect + 1) * blkdev->file_blk;
	page = vm_ref2ptr(blkdev->xendev.domain, ioreq->req.seg[i].gref);
	if (NULL == page) {
	    d1printf("%s: can't dereference grant\n", __FUNCTION__);
	    goto err;
	}
	mem = page + ioreq->req.seg[i].first_sect * blkdev->file_blk;
	d2printf("%s:   %d: op %d, mem %p, len %zd\n",
		 __FUNCTION__, i, ioreq->req.operation, mem, len);
	ioreq->vec[i].iov_base = mem;
	ioreq->vec[i].iov_len  = len;
	ioreq->end += len;
    }
    if (ioreq->end > blkdev->file_size) {
	d1printf("%s: access beyond end of file\n", __FUNCTION__);
	goto err;
    }
    ioreq->vecs = i;
    return 0;

err:
    blkdev->req_err_parse++;
    ioreq->status = BLKIF_RSP_ERROR;
    return -1;
}

/*
 * Run I/O using readv / writev.
 * Not thread safe due to lseek usage.
 */
static int ioreq_runio_serial(struct ioreq *ioreq)
{
    struct blkdev *blkdev = ioreq->blkdev;
    int len;
    
    if (ioreq->presync)
	fdatasync(blkdev->file);
	
    lseek(blkdev->file, ioreq->start, SEEK_SET);
    switch (ioreq->req.operation) {
    case BLKIF_OP_READ:
	len = readv(blkdev->file, ioreq->vec, ioreq->vecs);
	break;
    case BLKIF_OP_WRITE:
    case BLKIF_OP_WRITE_BARRIER:
	len = writev(blkdev->file, ioreq->vec, ioreq->vecs);
	break;
    default:
	goto err; /* unknown operation */
    }
    if (len != ioreq->end - ioreq->start)
	goto err; /* I/O error */

    if (ioreq->postsync)
	fdatasync(blkdev->file);

    if (BLKIF_OP_READ == ioreq->req.operation)
	blkdev->bytes_read += len;
    else
	blkdev->bytes_write += len;
    ioreq->status = BLKIF_RSP_OKAY;
    return 0;

err:
    blkdev->req_err_io++;
    ioreq->status = BLKIF_RSP_ERROR;
    return -1;
}

/*
 * Run I/O using pread / pwrite.
 * Needed for threaded I/O.
 */
static int ioreq_runio_parallel(struct ioreq *ioreq)
{
    struct blkdev *blkdev = ioreq->blkdev;
    int i, rc, len = 0;
    off_t pos;
    
    if (ioreq->presync)
	fdatasync(blkdev->file);
    switch (ioreq->req.operation) {
    case BLKIF_OP_READ:
	pos = ioreq->start;
	for (i = 0; i < ioreq->vecs; i++) {
	    rc = pread(blkdev->file,
		       ioreq->vec[i].iov_base,
		       ioreq->vec[i].iov_len,
		       pos);
	    if (rc != ioreq->vec[i].iov_len)
		goto err;
	    len += rc;
	    pos += rc;
	}
	break;
    case BLKIF_OP_WRITE:
    case BLKIF_OP_WRITE_BARRIER:
	pos = ioreq->start;
	for (i = 0; i < ioreq->vecs; i++) {
	    rc = pwrite(blkdev->file,
			ioreq->vec[i].iov_base,
			ioreq->vec[i].iov_len,
			pos);
	    if (rc != ioreq->vec[i].iov_len)
		goto err;
	    len += rc;
	    pos += rc;
	}
	break;
    default:
	goto err; /* unknown operation */
    }
    if (ioreq->postsync)
	fdatasync(blkdev->file);

    if (BLKIF_OP_READ == ioreq->req.operation)
	blkdev->bytes_read += len;
    else
	blkdev->bytes_write += len;
    ioreq->status = BLKIF_RSP_OKAY;
    return 0;

err:
    blkdev->req_err_io++;
    ioreq->status = BLKIF_RSP_ERROR;
    return -1;
}

static int blk_send_response_one(struct ioreq *ioreq)
{
    struct blkdev     *blkdev = ioreq->blkdev;
    int               send_notify   = 0;
    int               have_requests = 0;
    blkif_response_t  resp;
    void              *dst;
    
    resp.id        = ioreq->req.id;
    resp.operation = ioreq->req.operation;
    resp.status    = ioreq->status;
    
    /* Place on the response ring for the relevant domain. */
    switch (blkdev->protocol) {
    case BLKIF_PROTOCOL_NATIVE:
	dst = RING_GET_RESPONSE(&blkdev->rings.native, blkdev->rings.native.rsp_prod_pvt);
	break;
    case BLKIF_PROTOCOL_X86_32:
	dst = RING_GET_RESPONSE(&blkdev->rings.x86_32, blkdev->rings.x86_32.rsp_prod_pvt);
	break;
    case BLKIF_PROTOCOL_X86_64:
	dst = RING_GET_RESPONSE(&blkdev->rings.x86_64, blkdev->rings.x86_64.rsp_prod_pvt);
	break;
    }
    memcpy(dst, &resp, sizeof(resp));
    blkdev->rings.common.rsp_prod_pvt++;

    RING_PUSH_RESPONSES_AND_CHECK_NOTIFY(&blkdev->rings.common, send_notify);
    if (blkdev->rings.common.rsp_prod_pvt == blkdev->rings.common.req_cons) {
	/*
	 * Tail check for pending requests. Allows frontend to avoid
	 * notifications if requests are already in flight (lower
	 * overheads and promotes batching).
	 */
	RING_FINAL_CHECK_FOR_REQUESTS(&blkdev->rings.common, have_requests);
    } else if (RING_HAS_UNCONSUMED_REQUESTS(&blkdev->rings.common)) {
	have_requests = 1;
    }

    d2printf("%s: send_notify %d, have_requests %d\n",
	     __FUNCTION__, send_notify, have_requests);

    if (have_requests)
	blkdev->more_work++;
    return send_notify;
}

/* walk finished list, send outstanding responses, free requests */
static void blk_send_response_all(struct blkdev *blkdev)
{
    struct list_head *item, *safe;
    struct ioreq *ioreq;
    int send_notify = 0;

    pthread_mutex_lock(&blkdev->listlock);
    list_for_each_safe(item, safe, &blkdev->finished) {
	ioreq = list_entry(item, struct ioreq, list);
	send_notify += blk_send_response_one(ioreq);
	__ioreq_release(ioreq);
    }
    pthread_mutex_unlock(&blkdev->listlock);
    if (send_notify)
	xenner_evtchn_notify(evtchnd, blkdev->xendev.local_port);
}
	
static int blk_get_request(struct blkdev *blkdev, struct ioreq *ioreq, RING_IDX rc)
{
    switch (blkdev->protocol) {
    case BLKIF_PROTOCOL_NATIVE:
	memcpy(&ioreq->req, RING_GET_REQUEST(&blkdev->rings.native, rc),
	       sizeof(ioreq->req));
	break;
    case BLKIF_PROTOCOL_X86_32:
	blkif_get_x86_32_req(&ioreq->req, RING_GET_REQUEST(&blkdev->rings.x86_32, rc));
	break;
    case BLKIF_PROTOCOL_X86_64:
	blkif_get_x86_64_req(&ioreq->req, RING_GET_REQUEST(&blkdev->rings.x86_64, rc));
	break;
    }
    return 0;
}

static void blk_thread_main_wakeup(struct blkdev *blkdev)
{
    if (!blkdev->mainthread)
	return;
    pthread_mutex_lock(&blkdev->waitlock);
    blkdev->more_work++;
    pthread_cond_signal(&blkdev->waiter);
    pthread_mutex_unlock(&blkdev->waitlock);
}

static void blk_thread_main_gosleep(struct blkdev *blkdev)
{
    pthread_mutex_lock(&blkdev->waitlock);
    if (!blkdev->more_work)
	pthread_cond_wait(&blkdev->waiter, &blkdev->waitlock);
    blkdev->more_work = 0;
    pthread_mutex_unlock(&blkdev->waitlock);
}

static void *blk_thread_read_one(void *arg)
{
    struct ioreq *ioreq = arg;

    ioreq_runio_parallel(ioreq);
    ioreq_finish(ioreq);
    blk_thread_main_wakeup(ioreq->blkdev);
    return NULL;
}

static void *blk_thread_main(void *arg)
{
    struct blkdev *blkdev = arg;
    RING_IDX rc, rp;
    struct ioreq *ioreq;

    d1printf("%s: start\n", __FUNCTION__);
    for (;;) {
	if (blkdev->state != BLKIF_CONNECTED)
	    break;
	rc = blkdev->rings.common.req_cons;
	rp = blkdev->rings.common.sring->req_prod;
	rmb(); /* Ensure we see queued requests up to 'rp'. */
	
	while ((rc != rp)) {
	    /* pull request from ring */
	    if (RING_REQUEST_CONS_OVERFLOW(&blkdev->rings.common, rc))
		break;
	    ioreq = ioreq_start(blkdev);
	    if (NULL == ioreq)
		break;
	    blk_get_request(blkdev, ioreq, rc);
	    blkdev->rings.common.req_cons = ++rc;

	    /* parse them */
	    if (0 != ioreq_parse(ioreq)) {
		if (blk_send_response_one(ioreq))
		    xenner_evtchn_notify(evtchnd, blkdev->xendev.local_port);
		ioreq_release(ioreq);
		continue;
	    }

	    if (use_threads) {
		/* run i/o in threads mode */
		if (ioreq->req.operation == BLKIF_OP_READ) {
		    /* reads go async using a thread */
		    pthread_t thread;
		    pthread_create(&thread, &detached,
				   blk_thread_read_one, ioreq);
		} else {
		    /* writes are kept in order */
		    ioreq_runio_parallel(ioreq);
		    ioreq_finish(ioreq);
		}
	    } else {
		/* run i/o in sequential mode */
		ioreq_runio_serial(ioreq);
		ioreq_finish(ioreq);
	    }
	}
	blk_send_response_all(blkdev);
	blk_thread_main_gosleep(blkdev);
    }

    d1printf("%s: cleanup\n", __FUNCTION__);
    while (!list_empty(&blkdev->inflight))
	blk_thread_main_gosleep(blkdev);
    blk_send_response_all(blkdev);
    
    d1printf("%s: exit\n", __FUNCTION__);
    return NULL;
}

/* ------------------------------------------------------------- */

static int open_as(uid_t uid, gid_t gid, const char *file, int flags)
{
    int fd = -1, err = 0;
    uid_t ruid;
    gid_t rgid;

    ruid = getuid();
    rgid = getgid();

    if (rgid != gid  &&  -1 == setegid(gid)) {
	d1printf("setegid(%d): %s\n", gid, strerror(errno));
	goto fail_id;
    }
    if (ruid != uid  &&  -1 == seteuid(uid)) {
	d1printf("seteuid(%d): %s\n", uid, strerror(errno));
	goto fail_id;
    }

    fd = open(file, flags);
    if (-1 == fd) {
	err = errno;
	d1printf("open_as(%d, %d, %s): %s\n", uid, gid, file, strerror(errno));
    }

    if (ruid != uid  &&  -1 == seteuid(ruid)) {
	d1printf("seteuid(%d): %s\n", ruid, strerror(errno));
	goto fail_id;
    }
    if (rgid != gid  &&  -1 == setegid(rgid)) {
	d1printf("setegid(%d): %s\n", rgid, strerror(errno));
	goto fail_id;
    }

    errno = err;
    return fd;

fail_id:
    if (-1 != fd)
	close(fd);
    errno = EPERM;
    return -1;
}

static void blk_setup_backend(struct blkdev *blkdev)
{
    int mode, info = 0;
    
    if (blkdev->state != BLKIF_PROBED)
	return;
    
    /* read xenstore entries */
    if (NULL == blkdev->params)
	blkdev->params = read_be_str(&blkdev->xendev, "params");
    if (NULL == blkdev->mode)
	blkdev->mode = read_be_str(&blkdev->xendev, "mode");

    /* do we have all we need? */
    if (NULL == blkdev->params ||
	NULL == blkdev->mode)
	return;

    /* try opening file */
    if (0 == strcmp(blkdev->mode, "w")) {
	mode = O_RDWR;
    } else {
	mode = O_RDONLY;
	info |= VDISK_READONLY | VDISK_REMOVABLE | VDISK_CDROM;
    }
    if (insecure)
	blkdev->file = open(blkdev->params, mode);
    else
	blkdev->file = open_as(blkdev->xendev.domain->uid,
			       blkdev->xendev.domain->gid,
			       blkdev->params, mode);
    if (-1 == blkdev->file) {
	d1printf("%s: open %s: %s\n", __FUNCTION__,
		 blkdev->params, strerror(errno));
	return;
    }
    
    /* fill info */
    blkdev->file_blk  = BLOCK_SIZE;
    blkdev->file_size = lseek(blkdev->file, 0, SEEK_END);
    write_be_int(&blkdev->xendev, "feature-barrier", 1);
    write_be_int(&blkdev->xendev, "info",            info);
    write_be_int(&blkdev->xendev, "sector-size",     blkdev->file_blk);
    write_be_int(&blkdev->xendev, "sectors",
		 blkdev->file_size / blkdev->file_blk);

    change_state_xendev(&blkdev->xendev, XenbusStateInitWait);
    blkdev->state = BLKIF_INITIALISED;
}

static void blk_connect_frontend(struct blkdev *blkdev)
{
    void *sring;

    if (blkdev->state != BLKIF_INITIALISED)
	return;

    blkdev->ring_ref = read_fe_int(&blkdev->xendev, "ring-ref");
    blkdev->xendev.remote_port = read_fe_int(&blkdev->xendev, "event-channel");
    blkdev->xendev.protocol = read_fe_str(&blkdev->xendev, "protocol");
    if (0 == blkdev->ring_ref ||
	0 == blkdev->xendev.remote_port ||
	NULL == blkdev->xendev.protocol)
	return;

    blkdev->protocol = BLKIF_PROTOCOL_NATIVE;
    if (0 == strcmp(blkdev->xendev.protocol, XEN_IO_PROTO_ABI_X86_32))
	blkdev->protocol = BLKIF_PROTOCOL_X86_32;
    if (0 == strcmp(blkdev->xendev.protocol, XEN_IO_PROTO_ABI_X86_64))
	blkdev->protocol = BLKIF_PROTOCOL_X86_64;

    sring = vm_ref2ptr(blkdev->xendev.domain, blkdev->ring_ref);
    if (!sring)
	return;

    switch (blkdev->protocol) {
    case BLKIF_PROTOCOL_NATIVE:
    {
	blkif_sring_t *sring_native = sring;
	BACK_RING_INIT(&blkdev->rings.native, sring_native, PAGE_SIZE);
	break;
    }
    case BLKIF_PROTOCOL_X86_32:
    {
	blkif_x86_32_sring_t *sring_x86_32 = sring;
	BACK_RING_INIT(&blkdev->rings.x86_32, sring_x86_32, PAGE_SIZE);
	break;
    }
    case BLKIF_PROTOCOL_X86_64:
    {
	blkif_x86_64_sring_t *sring_x86_64 = sring;
	BACK_RING_INIT(&blkdev->rings.x86_64, sring_x86_64, PAGE_SIZE);
	break;
    }
    }

    blkdev->xendev.local_port = xenner_evtchn_bind_interdomain
	(evtchnd, blkdev->xendev.dom, blkdev->xendev.remote_port);

    d1printf("%s: ok: proto %s, ring-ref %d, "
	     "remote port %d, local port %d\n", __FUNCTION__,
	     blkdev->xendev.protocol, blkdev->ring_ref,
	     blkdev->xendev.remote_port, blkdev->xendev.local_port);

    change_state_xendev(&blkdev->xendev, XenbusStateConnected);
    blkdev->state = BLKIF_CONNECTED;

    pthread_create(&blkdev->mainthread, &joinable,
		   blk_thread_main, blkdev);
}

static void blk_disconnect(struct blkdev *blkdev, enum xenbus_state state)
{
    if (-1 != blkdev->file) {
	close(blkdev->file);
	blkdev->file = -1;
    }
    if (blkdev->xendev.local_port) {
	xenner_evtchn_unbind(evtchnd, blkdev->xendev.local_port);
	blkdev->xendev.local_port = 0;
    }

    if (state)
	change_state_xendev(&blkdev->xendev, state);
    blkdev->state = BLKIF_DISCONNECTED;
    blk_thread_main_wakeup(blkdev);
}

static void blk_print_stats(struct blkdev *blkdev, FILE *fp)
{
    fprintf(fp,
	    "stats dev: %d/%d\n"
	    "    read : bytes %" PRId64 " requests %d\n"
	    "    write: bytes %" PRId64 " requests %d barrier %d\n"
	    "    error: parse %d io %d\n"
	    "    queue: max %d/%d full %d\n"
	    "---\n",
	    blkdev->xendev.dom, blkdev->xendev.dev,
	    blkdev->bytes_read, blkdev->req_read,
	    blkdev->bytes_write, blkdev->req_write, blkdev->req_barrier,
	    blkdev->req_err_parse, blkdev->req_err_io,
	    blkdev->requests, max_requests, blkdev->req_full);
}

static void blk_check_ring(struct blkdev *blkdev)
{
    if (blkdev->state != BLKIF_CONNECTED)
	return;
    blk_thread_main_wakeup(blkdev);
}

static int blk_alloc(struct xendev *xendev)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    d1printf("%s: %p\n", __FUNCTION__, blkdev);
    blkdev->state = BLKIF_PROBED;

    INIT_LIST_HEAD(&blkdev->inflight);
    INIT_LIST_HEAD(&blkdev->finished);
    INIT_LIST_HEAD(&blkdev->freelist);
    return 0;
}

static int blk_setup(struct xendev *xendev)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    d1printf("%s: %p\n", __FUNCTION__, blkdev);
    blk_setup_backend(blkdev);
    blk_connect_frontend(blkdev);
    blk_check_ring(blkdev);
    return 0;
}

static int blk_backend(struct xendev *xendev, char *node, char *val)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    d2printf("%s: %p \"%s\" = \"%s\"\n", __FUNCTION__, blkdev, node,
	     val ? val : "<deleted>");
    blk_setup_backend(blkdev);
    return 0;
}

static int blk_frontend(struct xendev *xendev, char *node, char *val)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    d2printf("%s: %p \"%s\" = \"%s\"\n", __FUNCTION__, blkdev, node,
	     val ? val : "<deleted>");
    switch (xendev->fe_state) {
    case XenbusStateInitialised:
    case XenbusStateConnected:
	blk_connect_frontend(blkdev);
	blk_check_ring(blkdev);
	break;

    case XenbusStateClosing:
    case XenbusStateClosed:
	blk_disconnect(blkdev, xendev->fe_state);
	break;

    case XenbusStateInitialising:
    case XenbusStateInitWait:
    case XenbusStateUnknown:
	/* make gcc happy */
	break;
    }
    return 0;
}

static int blk_event(struct xendev *xendev)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);

    blk_check_ring(blkdev);
    return 0;
}

static int blk_free(struct xendev *xendev)
{
    struct blkdev *blkdev = container_of(xendev, struct blkdev, xendev);
    struct list_head *item, *safe;
    struct ioreq *ioreq;
    void *dummy;

    d1printf("%s: %p\n", __FUNCTION__, blkdev);

    if (debug)
	blk_print_stats(blkdev, stderr);
    blk_disconnect(blkdev, 0);
    if (blkdev->mainthread)
	pthread_join(blkdev->mainthread, &dummy);

    list_for_each_safe(item, safe, &blkdev->freelist) {
	ioreq = list_entry(item, struct ioreq, list);
	list_del(&ioreq->list);
	free(ioreq);
    }
	
    if (blkdev->params)
	free(blkdev->params);
    if (blkdev->mode)
	free(blkdev->mode);
    return 0;
}

static struct devops blkdev_ops = {
    .size  = sizeof(struct blkdev),
    .alloc = blk_alloc,
    .setup = blk_setup,
    .xs_be = blk_backend,
    .xs_fe = blk_frontend,
    .event = blk_event,
    .free  = blk_free,
};

/* ------------------------------------------------------------- */

static void usage(FILE *fp)
{
    fprintf(fp,
	    "\n"
	    "blkbackd --  xenner block backend daemon\n"
	    "\n"
	    "usage: blkbackd [options]\n"
	    "options:\n"
	    "   -h            print this text\n"
	    "   -d            increase debuglevel\n"
	    "   -p <file>     specify pidfile\n"
	    "\n"
	    "   -s            do syncronous writes\n"
	    "   -i            insecure mode\n"
	    "   -t            don't use threads (for async i/o)\n"
	    "   -m <nr>       max nr of requests (per dev)  [%d]\n"
	    "\n"
	    "-- \n"
	    "(c) 2007 Gerd Hoffmann <kraxel@redhat.com>\n"
	    "\n",
	    max_requests);
	    
}

int main(int argc, char *argv[])
{
    int c;

    for (;;) {
        if (-1 == (c = getopt(argc, argv, "hdistTp:m:")))
            break;
        switch (c) {
        case 'd':
	    debug++;
	    break;
	case 'i':
	    insecure = 1;
	    break;
	case 's':
	    syncwrite = 1;
	    break;
	case 't':
	    use_threads = 0;
	    break;
	case 'T':
	    use_threads = 1;
	    break;

	case 'p':
	    pidfile = optarg;
	    break;
	case 'm':
	    max_requests = atoi(optarg);
	    break;

        case 'h':
            usage(stdout);
            exit(0);
        default:
            usage(stderr);
            exit(1);
        }
    }

    pthread_attr_init(&joinable);
    pthread_attr_init(&detached);
    pthread_attr_setdetachstate(&joinable, PTHREAD_CREATE_JOINABLE);
    pthread_attr_setdetachstate(&detached, PTHREAD_CREATE_DETACHED);

    return mainloop(BE_ROOT, &blkdev_ops);
}
